!==================================================
!CONSTRAINED EFFICIENCY IN THE NEOCLASSICAL GROWTH
!MODEL WITH UNINSURABLE IDIOSYNCRATIC SHOCKS
!
!by Davila, Hong, Krusell, Rios-Rull
!
!
! Program to calculate the market outcome and
! planner's  steady state
!==================================================

program CONSTEFF

  implicit none

  ! global variable
  integer:: agrid, nshock, i, n, model, mmax, &
       iterflag=0, brolillas=0, update_flag=0, &
       pl_flag=0, m2direct
  real(8):: amin, amax, beta, AA, theta, sig, delta, aggk, aggl, rstart, rmin, rmax
  real(8):: r, w, aproxdr, aproxdis, yy, plterm, plterm0, plbnd(2)
  real(8):: update=1., update_pl=0.35, err=1.e-7
  real(8),parameter:: tinies=0.000001,zero=0.0, one=1.0
  real(8), dimension(:), allocatable:: a, e, mu, mgrid
  real(8), dimension(:,:),allocatable:: y0, c0, &
       ystar, cstar, pi, dist, fdist, fdistp, distemp


  integer:: in=10,out=20 ! File Unit
  character(13)::in_name="input.txt",out_name="output2.txt"
  
  !******* time keeper *******
  integer:: time_start, time_end, time_rate
  integer:: time_tot_start, time_tot_end
  integer, dimension(15):: time_cumu
  integer, parameter:: id=0
  
  !------------------------------------------------------------------
  !                      M A I N   P R O G R A M
  !------------------------------------------------------------------

  call time_keeper(0,0)

  ! Open main input and output files and copy input contents to output
  call time_keeper(1,1)
  call open_copy_output

  ! Read input file, initialize some variables, allocate some arrays
  ! and open some files
  call initialize
  call time_keeper(2,1)

  if (m2direct/=1) then
     model=1
     call aiyagari_eq   ! get mkt economy
     call time_keeper(1,5)
     write(6,*)"FINISHED ITERATION FOR k"
     write(7,*)"FINISHED ITERATION FOR k"
     write(6,*)"RESULTS FOR MARKET ECONOMY"
     write(7,*)"RESULTS FOR MARKET ECONOMY"
     write(out,*)"RESULTS FOR MARKET ECONOMY"
     call time_keeper(2,5)
     call time_keeper(-1,out)
     close(out)
  else
     brolillas=0
     model=2
     call aiyagari_eq    ! solve planner's problem
     call time_keeper(1,5)
     write(6,*)"FINISHED ITERATION FOR k"
     write(7,*)"FINISHED ITERATION FOR k"
     write(6,*)"RESULTS FOR PLANNERS ECONOMY"
     write(7,*)"RESULTS FOR PLANNERS ECONOMY"
     write(out,*)"RESULTS FOR PLANNERS ECONOMY"
     call time_keeper(2,5)
     call time_keeper(-1,out)
     close(out)
  end if

contains

  !=================================================================
  ! Subroutine aiyagari_eq
  !
  ! get steady state of aiyagari economy
  !=================================================================

  SUBROUTINE aiyagari_eq
    implicit none
    integer:: equil_iter, no_converge, no_converge_f(nshock,mmax)
    real (8) ::  k, k1, dk, rs(2), error
    logical :: dist_conv

    call time_keeper(1,4)

    dist_conv = .false.

    ! Solve for decision rules and asset distribution
    !  until asset distribution converges

    ! Rent and wage given capital-labor ratio


    rs(1)=1.041d0
    rs(2)=1.0d0

    !initial guess
    if (model==1) then
       r=rstart
       k=((r+delta-1.)/(AA*theta))**(1/(theta-1.))
!       k=7.75   ! guess for k/l ratio
    elseif (model==2) then
	   rs(1)=rmax
	   rs(2)=rmin
       if (m2direct==1) r=rstart
       k=((r+delta-1.)/(AA*theta))**(1/(theta-1.))
    endif

    do i=1,mmax
       if (mgrid(i)>k) exit
    end do

    do n=1,nshock
       dist(n,:)=mu(n)
    end do
    dist(:,1:i-1)=0.0

    w=fl(k,one)
    r=fk(k,one)+1.-delta

    ! Solve Decision Rules and Get Fixed Asset Distribution
    equil_iter = 0
    call time_keeper(2,4)

    do while (.not. dist_conv)
       equil_iter = equil_iter + 1
       write(6,*) 
       write(6,101) equil_iter,r,w
       write(7,*) 
       write(7,101) equil_iter,r,w
       write(out,101) equil_iter,r,w
101    format ('eqm it num=',i4,'   r=',f10.7,'   w=',f10.7)

       call time_keeper(1,2)
	   if (brolillas==0) then
	      c0=0.0;cstar=0.0
              y0=0.0;ystar=0.0
	   end if
       call decision_rules
       write(6,*) 
       write(6,*) 'FINISHED DECISION RULES and...'
       write(6,*) 
       write(7,*) 
       write(7,*) 'FINISHED DECISION RULES and...'
       write(7,*) 
       call time_keeper(2,2)


       call time_keeper(1,3)
       call sta_dist(fdistp)
       write(6,*)
       write(6,*)'FINISHED STA. DIST and...'
       write(6,*)
       write(7,*)
       write(7,*)'FINISHED STA. DIST and...'
       write(7,*)
       call time_keeper(2,3)


       call time_keeper(1,4)
       write(6,*) 
       write(6,*) 'FINISHED ITERATIONS AND NOW CALCULATING STATISTICS'
       write(7,*) 
       write(7,*) 'FINISHED ITERATIONS AND NOW CALCULATING STATISTICS'

       !Calculate More Aggregated Mean and Standard Deviations

       call getmean(fdistp,aggk,aggl)

       k1 = aggk/aggl

       ! Difference between previous and current capital-labor ratio

       dk = k - k1
       write(6,*)'  k/l ratio (k0) is     ',k
       write(7,*)'  k/l ratio (k0) is     ',k
       write(out,*)'  k/l ratio (k0) is     ',k
       write(6,*)'  k/l ratio (k1) is     ',k1
       write(7,*)'  k/l ratio (k1) is     ',k1
       write(out,*)'  k/l ratio (k1) is     ',k1

       if (model==1 .or. (model==2  .and. update_flag==1)) then

          write(6,*)'  aggk (supply)         ',aggk
          write(6,*)'  agg labor (l0) is     ',aggl
          write(6,*)'  dk (k-k1) is          ',dk
          write(6,*)'  plterm                ',plterm
          write(6,100)
          
          write(7,*)'  aggk (supply)         ',aggk
          write(7,*)'  agg labor (l0) is     ',aggl
          write(7,*)'  dk (k-k1) is          ',dk
          write(7,*)'  plterm                ',plterm
          write(7,100)
          
          write(out,*)'  aggk (supply)         ',aggk
          write(out,*)'  agg labor (l0) is     ',aggl
          write(out,*)'  dk (k-k1) is          ',dk
          write(out,*)'  plterm                ',plterm
          write(out,100)

100       format(1x,50('-'))



          !*************************************************************
          ! Check convergence
          !*************************************************************
          error=0.
          where (abs(fdistp-fdist)>aproxdis)
             no_converge_F=1
          elsewhere
             no_converge_F=0
          end where
          
          no_converge = sum(no_converge_F)
          error=maxval(abs(fdistp-fdist))
          
          if (no_converge==0) then
             dist_conv=.true.
          else																				  
             dist_conv=.false.
          end if
          
          if (100.0*abs(rs(1)-rs(2))/maxval(rs)<aproxdis) dist_conv=.true.
          ! updating


          brolillas=0
          update_flag=0
          
          
          if (dk<=0.) rs(1)=r
          if (dk>0.) rs(2)=r
          write(6,*) "current r", r
          write(7,*) "current r", r
          write(out,*) "current r", r
          if (equil_iter==1) then
             r = fk(k1,one) + (1.-delta)
             if (dk<=0.0 .and.r<rs(2)) then
                r=rs(2)
             elseif (dk>0.0 .and. r>rs(1)) then
                r=rs(1)
             end if
          else
             r=sum(rs)/2.
          end if
          k=((r+delta-1.)/(AA*theta))**(1/(theta-1.))
          w=fl(((r+delta-1.)/(AA*theta))**(1/(theta-1.)),one)
          write(6,*) "updated r", r
          write(7,*) "updated r", r
          write(out,*) "updated r", r
          call time_keeper(2,4)
          fdist=fdistp
       end if
    end do

    call time_keeper(1,4)
    
    call decision_print
    call output(aggk,aggl,r,w)
    
    call time_keeper(2,4)
    
  END SUBROUTINE aiyagari_eq


  !=================================================================
  ! Subroutine initialize
  !
  ! read parameter values
  ! allocate arrays
  !=================================================================
  SUBROUTINE initialize
    implicit none
    integer:: nn, mmax1, mmax2, m1, m2, amax1, amax2
    real(8):: amintemp, amaxtemp, temp
	real(8),dimension(:),allocatable::agrid1,agrid2,agridt,mgrid1,mgrid2,mgridt

    ! Read Parameters
    read(in,*)
	read(in,*) m2direct
    read(in,*) amin
    read(in,*) amax
    read(in,*) agrid
    read(in,*) beta
    read(in,*) AA
    read(in,*) theta
    read(in,*) sig
    read(in,*) delta


    read(in,*) aproxdr
    read(in,*) mmax
    read(in,*) aproxdis
    read(in,*) rstart, rmin, rmax

    read(in,*) nshock
    allocate(e(nshock));e=0.0
    read(in,*) e

    allocate(pi(nshock,nshock)); pi=0.0
    do i=1,nshock
       read(in,*) pi(i,:)
    end do

    allocate(mu(nshock)); mu=0.0
    do i=1,nshock
       read(in,*) mu(i)
    end do

    amax1=agrid/2
    amax2=agrid-amax1
    allocate(agrid1(amax1),agrid2(amax2),agridt(agrid))
    
    do nn=1,amax1
       agrid1(nn)=amin+(amax-amin)*(nn-1)/(amax1-1)
    end do
    do nn=1,amax2
       agrid2(nn)=dsqrt(amax-amin)*(nn-1)/(amax2-1)
    end do
    agrid2=amin+agrid2**2.
    agrid2(amax2)=amax
    agrid2(1)=amin
    
    m1=1;m2=1
    
    do nn=1,amax1+amax2
       if (agrid1(m1)>agrid2(m2)) then
          agridt(nn)=agrid2(m2)
          m2=m2+1
       elseif (agrid1(m1)<agrid2(m2)) then
          agridt(nn)=agrid1(m1)
          m1=m1+1
       else
          agridt(nn)=agrid1(m1)
          m1=m1+1
          m2=m2+1
       end if
       agrid=nn
       if (m1>amax1 .or. m2>amax2) exit
    end do
    agridt(1)=amin
    agridt(agrid)=amax
    
    allocate(a(agrid));a=0.0
    a=agridt(1:agrid)
        
    ! Construct grids on asset space
    mmax1=mmax/2
    mmax2=mmax-mmax1
    allocate(mgrid1(mmax1),mgrid2(mmax2),mgridt(mmax))
    
    do nn=1,mmax1
       mgrid1(nn)=amin+(amax-amin)*(nn-1)/(mmax1-1)
    end do
    do nn=1,mmax2
       mgrid2(nn)=dsqrt(amax-amin)*(nn-1)/(mmax2-1)
    end do
    mgrid2=amin+mgrid2**2.
    mgrid2(mmax2)=amax
    mgrid2(1)=amin
    
    m1=1;m2=1
    
    do nn=1,mmax1+mmax2
       if (mgrid1(m1)>mgrid2(m2)) then
          mgridt(nn)=mgrid2(m2)
          m2=m2+1
       elseif (mgrid1(m1)<mgrid2(m2)) then
          mgridt(nn)=mgrid1(m1)
          m1=m1+1
       else
          mgridt(nn)=mgrid1(m1)
          m1=m1+1
          m2=m2+1
       end if
       mmax=nn
       if (m1>mmax1 .or. m2>mmax2) exit
    end do
    mgridt(1)=amin
    mgridt(mmax)=amax
    
    allocate(mgrid(mmax))
    mgrid=mgridt(1:mmax)
    print*, mgrid(1:2)
    print*, a(1:2)
    allocate(y0(agrid,nshock)); y0=0.0
    allocate(c0(agrid,nshock)); c0=0.0
    allocate(ystar(agrid,nshock)); ystar=0.0
    allocate(cstar(agrid,nshock)); cstar=0.0

    allocate(dist(nshock,mmax));    dist=0
    allocate(fdist(nshock,mmax));   fdist=0
    allocate(fdistp(nshock,mmax));  fdistp=0
    allocate(distemp(nshock,mmax)); distemp=0

    deallocate(mgrid1,mgrid2,mgridt)

  END SUBROUTINE initialize



  !=================================================================
  ! Subroutine decision_print
  !
  ! print decision rules
  !=================================================================
  SUBROUTINE DECISION_PRINT
    implicit none
    integer:: n,i

    if (model==1) open(21,file='kvisA.out')
    if (model==1) open(22,file='deciA.out')
    if (model==2) open(21,file='kvisA2.out')
    if (model==2) open(22,file='deciA2.out')
    write(21,200) (e(i),i=1,nshock)
    write(21,201) 
    do n=1,agrid
       write(21,202) a(n),(ystar(n,i),i=1,nshock),a(n)-ystar(n,nshock)
       write(22,*) (ystar(n,i),i=1,nshock)
    end do

200 format (2x,'k\e',5x,3f10.3, '   a-y(a|high)')
201 format (50('-'))
202 format (f8.3, 2x, 4f10.3)

  END SUBROUTINE DECISION_PRINT


  !=================================================================
  ! Subroutine output
  !
  ! print decision rules
  !=================================================================
  SUBROUTINE output(kk,ll,rr,ww)
    implicit none
    real(8):: kk,ll,rr,ww,yy

    yy=F(kk,ll)

    write(out,*)
    write(out,100)
    write(out,*)
    write(out,101) kk
    write(out,102) ll
    write(out,103) yy
    write(out,104) kk/yy
    write(out,105) kk/ll
    write(out,106) (rr-1.0)*100
    write(out,107) ww
    write(out,*)
    write(out,100)

100 format(1x,60('-'))
101 format(4x,'Aggregate Assets   ',f14.6)
102 format(4x,'Aggregate Labors   ',f14.6)
103 format(4x,'Output             ',f14.6)
104 format(4x,'K/Y                ',f14.6)
105 format(4x,'K/L                ',f14.6)
106 format(4x,'Interest Rate      ',f14.6,'%')
107 format(4x,'Effective Wage     ',f14.6)

  END SUBROUTINE output

  !--------------------------------------------------------------------
  !                                 UPAP
  ! Given (ap,ep) calculate marginal utility
  ! Called by subroutine DECISION_RULES.
  !
  !--------------------------------------------------------------------
  FUNCTION UPAP (n,ap,ee) RESULT (upapresult)

    integer:: n, ee
    real (kind=8):: ap, cons, upapresult
    upapresult=0.

    cons=a(n)*r+w*e(ee)-ap
    upapresult=Uc(cons)


  END FUNCTION UPAP

  !--------------------------------------------------------------------
  !                           DECISION_RULES
  ! Solve for decision rules.
  ! given all the parameter values, this subroutine will get
  ! ystar(n,i) cstar(n,i)
  !--------------------------------------------------------------------

  SUBROUTINE DECISION_RULES
    implicit none
    integer, dimension (agrid,nshock):: no_converge_y
    integer:: iter, no_converge
    integer, parameter:: diterm=5000
    logical:: rules_converge

    real (kind=8):: up0, up_tmp, error, yexp, errrel
    real (kind=8):: meru0, meru_tmp
    real (8) :: du(agrid+1),dusup,yy1
    integer:: np

    errrel = 3.e-30

    ! Initialization 1 : general
    rules_converge =.false.
    iter=0


    !*************************************************************
    ! First Get last term
    !*************************************************************
    if (iter>1) iterflag=1
    plterm0=plterm
    plterm=0.0
    if (brolillas.lt.1) plterm0=0.0d0 ! reset plterm
    if (model==2  .and. pl_flag .ge. 1 .and. brolillas .ge. 1) then
       call exterm(plterm)
       if (dabs(plterm-plterm0)<1.0d-6) update_flag=1
       if (dabs(plbnd(1)-plbnd(2))<1.0d-9) update_flag=1
       if (plterm.ge.plterm0) plbnd(1)=max(plterm0,plbnd(1))
       if (plterm.le.plterm0) plbnd(2)=min(plterm0,plbnd(2))
       write(out,'(3x,A,2f12.8)') 'plterm(old,new) ', plterm0, plterm
       write(out,'(3x,A,2f12.8)') 'plterm[min,max] ', plbnd
       write(6,'(3x,A,2f12.8)') 'plterm[min,max] ', plbnd
       write(7,'(3x,A,2f12.8)') 'plterm[min,max] ', plbnd

       if (brolillas .ge. 3) then
          plterm=update_pl*plterm+(1.0d0-update_pl)*plterm0
          if (plterm>plbnd(2) .or. plterm<plbnd(1)) &
               plterm=sum(plbnd)/2.
       else
          plterm=update_pl*plterm+(1.0d0-update_pl)*plterm0
       end if
       write(out,'(3x,A,2f10.7)') 'plterm          ', plterm
       write(6,*) plterm
       write(7,*) plterm
    elseif (model==2 .and. brolillas==0) then
       plbnd(1)=-1.
       plbnd(2)=1.
       plterm=plterm0
       write(out,'(3x,A,2f10.7)') 'plterm          ', plterm0
       write(6,*) plterm0
       write(7,*) plterm0
    end if
    
    pl_flag=pl_flag+1
    brolillas=brolillas+1

    ! Initialization 2: construct cstar from ystar
    ystar=0.0
    cstar=0.0
    y0=0.0
    c0=0.0
    do i=1,nshock
       do n=1, agrid
          cstar(n,i)=r*a(n)+w*e(i)-ystar(n,i)
       end do
    end do

    ! Initialization 3 : initial updating
    y0 = (one-update)*y0+update*ystar
    c0 = (one-update)*c0+update*cstar


    !****************************************************************
    ! Main Loop
    !****************************************************************
    dowhile: do while (.not. rules_converge)

       iter = iter + 1
       write(6, 201)iter,minval(ystar),maxval(ystar)
       write(7, 201)iter,minval(ystar),maxval(ystar)

201    format ('iter = ',i4,'  ystar = [',f14.6,',',f14.6,']')


       !*************************************************************
       ! Solve the decision rule
       !*************************************************************
       do i=1,nshock
          do n=1,agrid
             !*** find ystar ***
             do np=1,agrid+1
                if (np>agrid) exit
                if (upap(n,a(np),i) > 100./tinies) exit
                du(np)=-upap(n,a(np),i)+meru(a(np),i)+plterm
                if (du(np)<=0.) exit
             end do

             if (np>agrid) then
                yy1=a(agrid)
             elseif (upap(n,a(np),i) > 100./tinies) then
                if (np==1) then
                   yy1=a(1)
                else
                   yexp=a(n)*r+w*e(i) - tinies                
                   dusup=upap(n,yexp,i)-meru(yexp,i)-plterm
                   if ((yexp-a(np-1))>tinies .and. dusup>0.) then
                      yy1=&
                           zbrent(foc,a(np-1),yexp,errrel)
                   else
                      yy1=a(np-1)
                   end if
                end if
             else
                if (np==1) then
                   yy1=a(1)
                else
                   yy1=zbrent(foc,a(np-1),a(np),errrel)
                end if
             end if

             ystar(n,i) = yy1
             cstar(n,i) = r*a(n)+w*e(i)-ystar(n,i)
          end do
       end do

       !*************************************************************
       ! Check convergence
       !*************************************************************
       error=0.
       where (abs(ystar-y0)>aproxdr)
          no_converge_y=1
       elsewhere
          no_converge_y=0
       end where

       no_converge = sum(no_converge_y)
       error=maxval(abs(ystar-y0))

       if (no_converge==0) then
          rules_converge=.true.
       else																				  
          rules_converge=.false.
       end if

       !*************************************************************
       ! Update y0 and c0
       !*************************************************************
       y0=ystar
       c0=cstar

       !*************************************************************
       ! Output the result
       !*************************************************************
       write(6,103)plterm,no_converge,error
       write(7,103)plterm,no_converge,error
103    format (2x,' planner term= ',f15.8,' # to conv=',i6,&
            ' max_error=',f15.8)

    end do dowhile

    if (model==1) open(23,file='drprAe.txt')
    if (model==2) open(23,file='drprAe2.txt')
    write(23,109)r
    write(23,103)plterm,no_converge,error
    do i=1,agrid
       
       write(23,108) a(i), ystar(i,:), ystar(i,nshock)-a(i)
    end do
108 format(2x,<nshock+2>f10.4)
109 format(2x,'interest rate= ',f10.7)
    
    close(23)
    write(6, 104)aproxdr,error
    write(7, 104)aproxdr,error
104 format (1x,' iter finished with aproxdr=',&
         f11.8,' and max_error=',f11.8)
    
  END SUBROUTINE decision_rules



  !--------------------------------------------------------------------
  !                                 MERU
  ! Solve fornext period's marginal expected utility.
  ! Called by subroutine DECISION_RULES.
  ! Uses ystar,cstar tomorrow to compute the marginal utility
  ! of consumption tomorrow
  !--------------------------------------------------------------------
  FUNCTION MERU (ap,ee) RESULT (upt)
    integer:: ee, eet
    real (kind=8):: ap, upt
    real (kind=8), dimension (nshock):: upts
    upt=0.

    do eet=1,nshock
       ap=max(ap,amin)
       if (ap<=amax) then
          upts(eet)=getupp(ap,eet)
       else
          upts(eet)=supergetupp(ap,eet)
       end if
       upt=upt+pi(ee,eet)*upts(eet)
    end do
    upt=upt*beta*r


  END FUNCTION MERU



  !---------------------------------------------------------------
  !                            GETUPP*
  ! Gets the marginal expected utility
  ! of a household of type i,g,z with assets a
  !---------------------------------------------------------------
  FUNCTION GETUPP(aaaa,eeee) RESULT(upp)
    implicit none
    integer:: nnnn, eeee
    real (kind=8):: aaaa, upp, consleft, consright, consint

    do nnnn=2,agrid
       if (aaaa<=a(nnnn)) goto 10
    end do
    nnnn=agrid
    print*,'Increase upper bound of assets getupp',aaaa

10  consleft=c0(nnnn-1,eeee)
    consright=c0(nnnn,eeee)

    consint=interpol(a(nnnn-1),a(nnnn),consleft,consright,aaaa)

    upp=Uc(consint)

  END FUNCTION GETUPP


  !---------------------------------------------------------------
  !                            SUPERGETUPP*
  ! Gets the marginal expected utility
  ! of a household of type i,g,z with assets a
  ! in case a is greater than amax
  !---------------------------------------------------------------
  FUNCTION SUPERGETUPP(aaaa,eeee) RESULT(upp)
    implicit none
    integer:: nnnn, eeee
    real (kind=8):: aaaa, upp, uppamax, uppamaxminus
    real (kind=8):: slopeamax, upptemp

    if (aaaa<=amax) then
       aaaa=amax
       print*,'aaaa<=amax in SUPERGETUPP:',aaaa
    end if

    uppamax=getupp(amax,eeee)
    uppamaxminus=getupp(amax-tinies,eeee)

    slopeamax=(uppamax-uppamaxminus)/tinies

    upptemp=uppamax+slopeamax*(aaaa-amax)
    upp=max(zero,upptemp)

  END FUNCTION SUPERGETUPP



  !------------------------------------------------------------------
  !                                 GETSAV
  !  Gets the savings of a household of type i,g,z with assets a
  !------------------------------------------------------------------
  function getsav (aaaa,eeee) result(savings)

    integer :: nnnn, eeee
    real ( kind =8 ) :: aaaa, savings, normaaaa
    normaaaa=aaaa
    do nnnn = 2, agrid
       if ( normaaaa <= a(nnnn) ) goto 10
    end do
    nnnn=agrid
    write(6,*)'hits the upperbound of asset in GETSAV',&
         normaaaa
    write(7,*)'hits the upperbound of asset in GETSAV',&
         normaaaa
10  savings = interpol(a(nnnn-1), a(nnnn), &
         ystar(nnnn-1,eeee),ystar(nnnn,eeee), normaaaa)
  end function getsav


  function foc(ain)
    implicit none
    real(8):: ain, foc

    foc =  plterm+meru(ain,i)-upap(n,ain,i)

  end function foc


  !----------------------------------------------------------------
  !                             EXTERM
  !  Gets the extra term of the planner's problem
  !----------------------------------------------------------------
  subroutine exterm(value)
    implicit none
    real(8),intent(out)::value
    real(8)::lk,kk,al,ar,cl,cr,mcl,mcr
    integer:: m, s

    call time_keeper(2,2) ! pause decision time
    call time_keeper(1,3) ! resume distribution time

    value=0.0

    dist=fdistp
    
    call time_keeper(2,3) ! pause distribution time
    call time_keeper(1,4) ! resume stat time
    
    call getmean(dist,aggk,aggl)
    kk=fkk(aggk,aggl)
    lk=flk(aggk,aggl)
    do s=1,nshock
       do m=1,mmax-1
          al=mgrid(m)
          ar=mgrid(m+1)
          cl=r*al+w*e(s)-getsav(al,s)
          cr=r*ar+w*e(s)-getsav(ar,s)
          mcl=Uc(cl)
          mcr=Uc(cr)
          
          if (m==1) then
             value=value+dist(s,m)*&
                  (e(s)*lk+al*kk)*mcl
          end if
          
          value=value+(dist(s,m+1)-dist(s,m))/(ar-al)* &
               (e(s)*lk*(mcl*ar-mcr*al)+ &
               0.5*(al+ar)* &
               (e(s)*lk*(mcr-mcl)+kk*(mcl*ar-mcr*al))+&
               kk*(mcr-mcl)*(ar**2.+al**2.+ar*al)/3.)
       end do
    end do
    
    value=beta*value
    
    open(318,file="dist_misc.txt")
    write(318,319) 'AggK= ', aggk
    write(318,319) 'AggL= ', aggl
    write(318,319) 'val = ', value
    write(318,320) 'maxm= ', mmax
319 format(A,f15.10)
320 format(A,i8)
    close(318)


    call time_keeper(2,4) ! pause stat time
    call time_keeper(1,2) ! resume dicision time

  end subroutine exterm

  !=================================================================
  ! Subroutine sta_dist
  !
  ! get invariant distribution over asset and shock
  !=================================================================
  subroutine sta_dist(sdist)
    implicit none
    integer:: s, sp, m, mi, no_converge_s, dis_iter
    integer:: no_converge_st(nshock,mmax)
    real(8):: prv_a, error, Ftemp
    real(8),dimension(nshock,mmax):: F00, F11
    real(8),dimension(nshock,mmax),intent(out):: sdist
    logical:: dist_conv



    F00=0.0; F11=0.0
    dis_iter=0
    dist_conv=.false.
    sdist=0.0

    ! Initialize F0 
    ! All households hold the minimum wealth the first time and
    ! previous distribution other times

    do s=1,nshock
       F00(s,:)=mu(s)
    end do
    


    do while (.not. dist_conv)
       
       dis_iter=dis_iter+1
       
       do sp=1,nshock
          do m=1,mmax
             prv_a=invystar(sp,mgrid(m))
             if (prv_a<amin) then
                Ftemp=0.0
             elseif (prv_a>=amax) then
                Ftemp=F00(sp,mmax)
             else
                do mi=2,mmax
                   if (mgrid(mi)>prv_a) exit
                end do
                if (mi==mmax+1) then
                   Ftemp=F00(sp,mmax)
                endif
                Ftemp=interpol(mgrid(mi-1),mgrid(mi), &
                     F00(sp,mi-1),F00(sp,mi),prv_a)
             end if

             do s=1,nshock
                F11(s,m)=F11(s,m)+pi(sp,s)*Ftemp
             end do
          end do
       end do

       !*************************************************************
       ! Check convergence
       !*************************************************************
       error=0.
       where (abs(F11-F00)>aproxdis)
          no_converge_st=1
       elsewhere
          no_converge_st=0
       end where

       no_converge_s = sum(no_converge_st)
       error=maxval(abs(F11-F00))

       if (no_converge_s==0) then
          dist_conv=.true.
       else																				  
          dist_conv=.false.
       end if

       ! Updating
       F00=F11
       F11=0.0

       if (mod(dis_iter,10)==1) then
          write(6,304) dis_iter, no_converge_s, error, aproxdis
          write(7,304) dis_iter, no_converge_s, error, aproxdis
       endif

304    format (1x, 'dist_iter= ',i6,' # to conv = ', i5, ' max_error= ', f8.6, ' tol= ',f8.6)

    end do


    sdist=F00

    if (model==1) open(37,file='distA.txt') 
    if (model==2) open(37,file='distA2.txt')
    open(317,file="dist00.txt")
    write(37,*) mmax
    write(317,'(2f12.9)') r, plterm
    do m=1,mmax
       write(37,*) mgrid(m), F00(:,m)
       write(317,'(4f12.6)') mgrid(m), (F00(s,m), s=1,nshock)
    end do
    close(37)
    close(317)
  end subroutine sta_dist


  !=================================================================
  ! Function invystar
  !
  ! get inverse of decision rule
  !=================================================================
  real(8) function invystar(sh,ap)

    integer:: sh, ah
    real(8):: ap

    ! Find out which interval ap is in

    do ah=2,agrid
       if (ap<ystar(ah,sh)) exit
    end do

    if (ah==agrid+1) then
       invystar=amax
    elseif (ystar(ah-1,sh)==ystar(ah,sh)) then
       invystar=a(ah)
    else         
       invystar=interpol(ystar(ah-1,sh),ystar(ah,sh),a(ah-1),a(ah),ap)
    end if

  end function invystar

  !=================================================================
  ! subroutine getmean
  !
  ! get mean from distribution function
  !=================================================================
  subroutine getmean(fdis,tk,tl)
    integer::m
    real(8),intent(out)::tk, tl
    real(8):: fdis(nshock,mmax),temp(mmax)

    temp=sum(fdis,dim=1)

    tk=0.0
    tl=0.0
	tk=temp(1)*mgrid(1)
    do m=1,mmax-1
       tk=tk+0.5*(temp(m+1)-temp(m))*(mgrid(m)+mgrid(m+1))
!	   write(*,'(3f15.10)') temp(m),mgrid(m), tk
	   
    end do

    tl=sum(fdis(:,mmax)*e)

  end subroutine getmean








  !=================================================================
  ! Function Uc
  !
  ! get marginal utility of (input) consumption value
  !=================================================================
  FUNCTION Uc(cin)
    implicit none
    real(8):: cin,Uc

    if (cin<tinies) then
       Uc=500./tinies
    else
       Uc=cin**(-sig)
    end if

  END FUNCTION Uc

  !=================================================================
  ! Function Interpol
  !   linear interpolation of y(ax) between a0,a1
  !=================================================================
  FUNCTION INTERPOL (a0,a1,y0,y1,ax) RESULT (yint)
    implicit none
    real(8):: a0, a1, y0, y1, ax, yint
    yint=y0+(ax-a0)/(a1-a0)*(y1-y0)
  END FUNCTION INTERPOL

  !=================================================================
  ! Function F,Fk, Fl, Fkk, Flk
  !
  ! get output from production function
  !=================================================================
  FUNCTION F(kin,lin)
    implicit none
    real(8):: kin,lin,F

    F=AA*kin**theta*lin**(1.-theta)

  END FUNCTION F

  FUNCTION Fk(kin,lin)
    implicit none
    real(8):: kin,lin,Fk

    Fk=AA*theta*kin**(theta-1.)*lin**(1.-theta)

  END FUNCTION Fk

  FUNCTION Fl(kin,lin)
    implicit none
    real(8):: kin,lin,Fl

    Fl=AA*(1.-theta)*kin**theta*lin**(-theta)

  END FUNCTION Fl

  FUNCTION Fkk(kin,lin)
    implicit none
    real(8):: kin,lin,Fkk

    Fkk=AA*theta*(theta-1.)*kin**(theta-2.)*lin**(1.-theta)

  END FUNCTION Fkk

  FUNCTION Flk(kin,lin)
    implicit none
    real(8):: kin,lin,Flk

    Flk=AA*(1.-theta)*theta*kin**(theta-1.)*lin**(-theta)

  END FUNCTION Flk


  !********************************************************************
  ! function: zbrent
  ! from Numerical Recipes in Fortran 77, pp354-355.
  !********************************************************************
  FUNCTION zbrent (func,x1,x2,tol)
    implicit none
    real(8), external:: func
    real(8), intent(in):: x1,x2,tol
    integer, parameter:: itmax=300
    real(8), parameter:: epsz=3.e-8
    integer:: iter
    real(8):: a,b,c,d,e,fa,fb,fc,p,q,r,s,tol1,xm, zbrent

    a=x1
    b=x2
    fa=func(a)
    fb=func(b)
    if(fa*fb>0.) pause 'root must be bracketed for zbrent'
    c=b
    fc=fb
    do iter=1,itmax
       if (fb*fc>0.) then
          c=a
          fc=fa
          d=b-a
          e=d
       end if
       if (abs(fc)<abs(fb)) then
          a=b
          b=c
          c=a
          fa=fb
          fb=fc
          fc=fa
       end if
       tol1=2.*epsz*abs(b)+0.5*tol
       xm=.5*(c-b)
       if (abs(xm)<=tol1 .or. fb==0.) then
          zbrent=b
          return
       end if
       if (abs(e)>=tol1 .and. abs(fa)>abs(fb)) then
          s=fb/fa
          if(a==c) then
             p=2.*xm*s
             q=1.-s
          else
             q=fa/fc
             r=fb/fc
             p=s*(2.*xm*q*(q-r)-(b-a)*(r-1.))
             q=(q-1.)*(r-1.)*(s-1.)
          end if
          if(p>0.) q=-q
          p=abs(p)
          if (2.*p<min(3.*xm*q-abs(tol1*q),abs(e*q))) then
             e=d
             d=p/q
          else
             d=xm
             e=d
          end if
       else
          d=xm
          e=d
       end if
       a=b
       fa=fb
       if(abs(d)>tol1) then
          b=b+d
       else
          b=b+sign(tol1,xm)
       end if
       fb=func(b)
    end do
    pause 'zbrent exceeding maximum iterations'
    zbrent=b
    return
  END FUNCTION zbrent

  !=================================================================
  ! Subroutine open_copy_output
  !
  ! make a copy of input on the output file
  !=================================================================
  SUBROUTINE open_copy_output
    implicit none
    integer:: nline,nn
    character(72):: line

    open(in,file=in_name,status='old')
    read(in,*) nline
    rewind(unit=in)
    open(out,file=out_name)
    write(out,*) ' This is from ',in_name
    do nn=1,nline
       read(in,80) line
       write(out,81) line
    end do
    rewind(unit=in)
    write(out,*) 
    write(out,*) '----------------- The Output begins -------------'
    write(out,*) 

80  format(a72)
81  format(1x,a72)

    open(7,file="screen2.txt")

  END SUBROUTINE open_copy_output

  !********************************************************************
  ! subroutine: time_keeper
  ! created for mskshock2 in february 2002
  ! if use, declare the following as globals:
  ! integer:: time_start, time_end, time_rate
  ! integer:: time_tot_start, time_tot_end
  ! integer, dimension(15):: time_cumu
  ! id=identifier of the processor (0, if single)
  !********************************************************************
  SUBROUTINE time_keeper (time_mode1, time_mode2)
    implicit none
    integer, intent(in):: time_mode1, time_mode2

    if (time_mode1==0) then !initialization
       call system_clock(count=time_tot_start, count_rate=time_rate)
       if (time_rate==0) time_rate=1
       time_cumu(1:15)=0
    else if (time_mode1==1) then !set starting time
       call system_clock(count=time_start)
    else if (time_mode1==2) then !check time and record it
       call system_clock(count=time_end)
       time_cumu(time_mode2)=time_cumu(time_mode2)+&
            time_end-time_start
    else if (time_mode1==-1) then !finalization
       call system_clock(count=time_tot_end)
       time_cumu=time_cumu/time_rate
       if (id==0) then
          write (time_mode2, 500)
          write (time_mode2, 501)
          write (time_mode2, 510) (time_tot_end-time_tot_start)/time_rate
          write (time_mode2, 511) time_cumu(1)
          write (time_mode2, 512) time_cumu(2)
          write (time_mode2, 513) time_cumu(3)
          write (time_mode2, 514) time_cumu(4)
          write (time_mode2, 515) time_cumu(5)
          write (time_mode2, 500)
          write (time_mode2,*)
          write (time_mode2,*)
          write (time_mode2,*)

          print 500
          print 501
          print 510, (time_tot_end-time_tot_start)/time_rate
          print 511, time_cumu(1)
          print 512, time_cumu(2)
          print 513, time_cumu(3)
          print 514, time_cumu(4)
          print 515, time_cumu(5)
          print 500
          print *, ''
          print *, ''
          print *, ''

          write(7,500)
          write(7,501)
          write(7,510) (time_tot_end-time_tot_start)/time_rate
          write(7,511) time_cumu(1)
          write(7,512) time_cumu(2)
          write(7,513) time_cumu(3)
          write(7,514) time_cumu(4)
          write(7,515) time_cumu(5)
          write(7,500)
          write(7,*)
          write(7,*)
          write(7,*)
       end if
    end if

500 format ('--------------------------------------------------')
501 format ('REPORT OF TIME SPENT FOR EACH ROUTINE')
510 format ('  TOTAL TIME           : ',i8,' SECONDS')
511 format ('  INITIALIZATION       : ',i8,' SECONDS')
512 format ('  DECISION RULES       : ',i8,' SECONDS')
513 format ('  GET DISTRIBUTION     : ',i8,' SECONDS')
514 format ('  STATS CALCULATION,ETC: ',i8,' SECONDS')
515 format ('  CLOSING PROCEDURE    : ',i8,' SECONDS')

  END SUBROUTINE time_keeper


end program CONSTEFF
